using ImmichToSlideshow.Models;
using ImmichToSlideshow.Models.Immich;
using Microsoft.AspNetCore.Mvc;
using Npgsql;
using System.Collections.ObjectModel;
using System.Data;
using System.Text;
using System.Text.Json;

namespace ImmichToSlideshow.Services;

public class AssetService(ILogger<Program> logger, AppSettings appSettings)
{

    private readonly ILogger<Program> _Logger = logger;
    private readonly AppSettings _AppSettings = appSettings;

    private static string GetColumnsCommandText()
    { // cSpell:disable
        List<string> results = new();
        // results.Add(" SELECT COALESCE(SUM(checksum_failures), 0) ");
        // results.Add(" FROM pg_stat_database ");
        // results.Add(" SELECT json_agg(t) ");
        // results.Add("   FROM information_schema.tables t ");
        // results.Add("  WHERE table_schema='public' ");
        // results.Add("    AND table_type='BASE TABLE' ");
        results.Add(" SELECT json_agg(c) ");
        results.Add("   FROM information_schema.columns c ");
        // results.Add("  WHERE table_name ='assets' ");
        // results.Add("  WHERE table_name ='libraries' ");
        results.Add("  WHERE table_name ='asset_files' ");
        return string.Join(Environment.NewLine, results);
    } // cSpell:enable

    private static string GetOwnerIdActiveImageCommandText()
    { // cSpell:disable
        List<string> results = new();
        results.Add(" SELECT json_agg(j) ");
        results.Add(" FROM ( ");
        results.Add("  SELECT a.\"ownerId\" ");
        results.Add("    FROM assets a ");
        results.Add("   WHERE a.\"status\" = 'active' ");
        results.Add("     AND a.\"type\" = 'IMAGE' ");
        results.Add("   GROUP");
        results.Add("      BY a.\"ownerId\" ");
        results.Add(" ) j ");
        return string.Join(Environment.NewLine, results);
    } // cSpell:enable

    private static string GetAssetActiveImagePreviewNotDuplicateCommandText()
    { // cSpell:disable
        List<string> results = new();
        results.Add(" SELECT json_agg(j) ");
        results.Add(" FROM ( ");
        results.Add("  SELECT a.\"id\" ");
        results.Add("       , a.\"deviceAssetId\" ");
        // results.Add("       , a.\"ownerId\" ");
        // results.Add("       , a.\"deviceId\" ");
        // results.Add("       , a.\"type\" ");
        results.Add("       , a.\"originalPath\" ");
        // results.Add("       , a.\"fileCreatedAt\" ");
        // results.Add("       , a.\"fileModifiedAt\" ");
        // results.Add("       , a.\"isFavorite\" ");
        // results.Add("       , a.\"duration\" ");
        // results.Add("       , a.\"encodedVideoPath\" ");
        // results.Add("       , a.\"checksum\" ");
        // results.Add("       , a.\"isVisible\" ");
        // results.Add("       , a.\"livePhotoVideoId\" ");
        // results.Add("       , a.\"updatedAt\" ");
        // results.Add("       , a.\"createdAt\" ");
        // results.Add("       , a.\"isArchived\" ");
        results.Add("       , a.\"originalFileName\" ");
        // results.Add("       , a.\"sidecarPath\" ");
        // results.Add("       , a.\"thumbhash\" ");
        // results.Add("       , a.\"isOffline\" ");
        // results.Add("       , a.\"libraryId\" ");
        // results.Add("       , a.\"isExternal\" ");
        // results.Add("       , a.\"deletedAt\" ");
        // results.Add("       , a.\"localDateTime\" ");
        // results.Add("       , a.\"stackId\" ");
        results.Add("       , a.\"duplicateId\" ");
        // results.Add("       , a.\"status\" ");
        results.Add("       , f.\"path\" ");
        results.Add("    FROM assets a ");
        results.Add("   INNER ");
        results.Add("    JOIN asset_files f ");
        results.Add("      ON a.\"id\" = f.\"assetId\" ");
        results.Add("   WHERE a.\"status\" = 'active' ");
        results.Add("     AND a.\"type\" = 'IMAGE' ");
        results.Add("     AND f.\"type\" = 'preview' ");
        results.Add("     AND a.\"duplicateId\" is null ");
        results.Add("     AND a.\"isArchived\" = false ");
        results.Add("     AND a.\"isExternal\" = true ");
        results.Add("     AND a.\"isOffline\" = false ");
        results.Add("     AND a.\"isVisible\" = true ");
        results.Add("     AND a.\"ownerId\" = @ownerId ");
        results.Add(" ) j ");
        return string.Join(Environment.NewLine, results);
    } // cSpell:enable

    private static string SetAssetArchivedCommandText(string deviceAssetIds)
    { // cSpell:disable
        List<string> results = new();
        results.Add(" UPDATE assets ");
        results.Add("    SET \"isArchived\" = true ");
        results.Add("  WHERE \"isArchived\" = false ");
        results.Add("    AND \"type\" = 'IMAGE' ");
        results.Add("    AND \"ownerId\" = @ownerId ");
        results.Add("    AND \"deviceAssetId\" in ( ");
        results.Add(deviceAssetIds);
        results.Add("                             ) ");
        return string.Join(Environment.NewLine, results);
    } // cSpell:enable

    private static int? ExecuteNonQuery(string connectionString, string commandText, NpgsqlParameter[] npgsqlParameters)
    {
        int? result;
        if (string.IsNullOrEmpty(connectionString))
            result = null;
        else
        {
            using NpgsqlConnection npgsqlConnection = new(connectionString);
            npgsqlConnection.Open();
            using NpgsqlCommand npgsqlCommand = new(commandText, npgsqlConnection);
            npgsqlCommand.Parameters.AddRange(npgsqlParameters);
            result = npgsqlCommand.ExecuteNonQuery();
        }
        return result;
    }

    private static StringBuilder GetForJsonPath(string connectionString, string commandText, NpgsqlParameter[] npgsqlParameters)
    {
        StringBuilder result = new();
        using NpgsqlConnection npgsqlConnection = new(connectionString);
        npgsqlConnection.Open();
        using NpgsqlCommand npgsqlCommand = new(commandText, npgsqlConnection);
        npgsqlCommand.Parameters.AddRange(npgsqlParameters);
        NpgsqlDataReader npgsqlDataReader = npgsqlCommand.ExecuteReader(CommandBehavior.SequentialAccess);
        while (npgsqlDataReader.Read())
            _ = result.Append(npgsqlDataReader.GetString(0));
        return result;
    }

    public string? GetColumns()
    {
        string result;
        string commandText = GetColumnsCommandText();
        NpgsqlParameter[] npgsqlParameters = [];
        StringBuilder stringBuilder = GetForJsonPath(_AppSettings.ConnectionString, commandText, npgsqlParameters);
        result = stringBuilder.ToString();
        if (result.Length == 1)
            File.WriteAllText(".vscode/jsonl/.jsonl", result);
        return result;
    }

    public string? GetOwnerIds()
    {
        string result;
        string commandText = GetOwnerIdActiveImageCommandText();
        NpgsqlParameter[] npgsqlParameters = [];
        StringBuilder stringBuilder = GetForJsonPath(_AppSettings.ConnectionString, commandText, npgsqlParameters);
        result = stringBuilder.ToString();
        if (result.Length == 1)
            File.WriteAllText(".vscode/jsonl/.jsonl", result);
        return result;
    }

    public string? GetAssets(Guid ownerId)
    {
        string result;
        string commandText = GetAssetActiveImagePreviewNotDuplicateCommandText();
        NpgsqlParameter[] npgsqlParameters = [new NpgsqlParameter(nameof(ownerId), ownerId)];
        StringBuilder stringBuilder = GetForJsonPath(_AppSettings.ConnectionString, commandText, npgsqlParameters);
        result = stringBuilder.ToString();
        if (result.Length == 1)
            File.WriteAllText(".vscode/jsonl/assets.jsonl", result);
        return result;
    }

    public ReadOnlyCollection<string>? GetRandomPaths(Guid ownerId)
    {
        string[]? results;
        string commandText = GetAssetActiveImagePreviewNotDuplicateCommandText();
        NpgsqlParameter[] npgsqlParameters = [new NpgsqlParameter(nameof(ownerId), ownerId)];
        StringBuilder stringBuilder = GetForJsonPath(_AppSettings.ConnectionString, commandText, npgsqlParameters);
        string json = stringBuilder.ToString();
        Random random = new();
        string ownerIdValue = ownerId.ToString();
        Asset[]? assets = JsonSerializer.Deserialize(json, AssetCollectionSourceGenerationContext.Default.AssetArray);
        results = assets is null ? null : (from l in assets orderby random.NextSingle() select l.Path.Split(ownerIdValue)[1]).ToArray();
        return results?.AsReadOnly();
    }

    public string? SaveRandomPaths(Guid ownerId)
    {
        string? results;
        DateTime dateTime = DateTime.Now;
        string tomorrow = dateTime.AddDays(1).ToString("MM-dd");
        if (Directory.Exists(_AppSettings.RandomResultsDirectory))
            _ = Directory.CreateDirectory(_AppSettings.RandomResultsDirectory);
        FileInfo fileInfo = new(Path.Combine(_AppSettings.RandomResultsDirectory, $"{tomorrow}.json"));
        if (fileInfo.Exists && fileInfo.CreationTime > dateTime.AddDays(_AppSettings.AddDays))
            results = null;
        else
        {
            _Logger.LogDebug("Writing <{FullName}>", fileInfo.FullName);
            ReadOnlyCollection<string>? paths = GetRandomPaths(ownerId);
            if (paths is null)
                results = null;
            else
            {
                _Logger.LogInformation("{count} path(s)", paths.Count.ToString());
                results = JsonSerializer.Serialize(paths);
                File.WriteAllText(fileInfo.FullName, results);
            }
        }
        return results;
    }

    private record Record(string Source, string Destination);

    public ReadOnlyCollection<string> SyncImmich(Guid ownerId)
    {
        List<string> results = [];
        Record record;
        List<Record> records = [];
        if (Directory.Exists(_AppSettings.SyncDirectory))
            _ = Directory.CreateDirectory(_AppSettings.SyncDirectory);
        int syncLength = _AppSettings.SyncDirectory.Length;
        string[] syncFiles = Directory.GetFiles(_AppSettings.SyncDirectory, "*", SearchOption.AllDirectories);
        string[] syncCheck = syncFiles.Select(l => l[syncLength..]).ToArray();
        if (Directory.Exists(_AppSettings.ImmichUploadDirectory))
            _ = Directory.CreateDirectory(_AppSettings.ImmichUploadDirectory);
        int immichUploadLength = _AppSettings.ImmichUploadDirectory.Length;
        string[] immichUploadFiles = Directory.GetFiles(_AppSettings.ImmichUploadDirectory, "*", SearchOption.AllDirectories);
        string[] immichUploadCheck = immichUploadFiles.Select(l => l[immichUploadLength..]).ToArray();
        for (int i = 0; i < immichUploadFiles.Length; i++)
        {
            if (syncCheck.Contains(immichUploadCheck[i]))
                continue;
            results.Add(immichUploadCheck[i]);
            record = new(immichUploadFiles[i], string.Concat(_AppSettings.SyncDirectory, immichUploadCheck[i]));
            records.Add(record);
        }
        _Logger.LogInformation("{count} file(s)", results.Count.ToString());
        for (int i = 0; i < records.Count; i++)
        {
            record = records[i];
            _Logger.LogDebug("Copying <{source}>", record.Source);
            File.Copy(record.Source, record.Destination);
        }
        return results.AsReadOnly();
    }

    public ReadOnlyCollection<int> SetArchiveImmich(Guid ownerId)
    {
        List<int> results;
        string checkDirectory = Path.Combine(_AppSettings.CurrentResultsDirectory, "B)Metadata", _AppSettings.CurrentCommit, "[]");
        if (!Directory.Exists(checkDirectory))
            results = null;
        else
        {
            string checkFile = Path.Combine(checkDirectory, "!9.json");
            if (!File.Exists(checkFile))
                results = null;
            else
            {
                string json = File.ReadAllText(checkFile);
                Identifier[]? identifiers = JsonSerializer.Deserialize<Identifier[]>(json);
                if (identifiers is null || identifiers.Length == 0)
                    results = null;
                else
                {
                    string deviceAssetIds = $"'{string.Join($"',{Environment.NewLine}'", identifiers.Select(l => $"{l.PaddedId}{l.Extension}").ToArray())}'";
                    string commandText = SetAssetArchivedCommandText(deviceAssetIds);
                    NpgsqlParameter[] npgsqlParameters = [new NpgsqlParameter(nameof(ownerId), ownerId)];
                    int? result = ExecuteNonQuery(_AppSettings.ConnectionString, commandText, npgsqlParameters);
                    if (result is null)
                        results = null;
                    else
                        results = [result.Value];
                }
            }
        }
        return results?.AsReadOnly();
    }

}